@@ -31,13 +31,31 @@ so keep your hardhats on!  A good place to start is the `FileStore` class in the
 current code.
 """
 
-__version__ = '0.3.0'
-
-import os
-from os import path
+__version__ = '0.6.0'
 
 
-packagedir = path.dirname(path.abspath(__file__))
-assert path.isdir(packagedir)
-datadir = path.join(packagedir, 'data')
-assert path.isdir(datadir)
+def configure_logging(namespace):
+    import os
+    from os import path
+    import logging
+
+    import xdg.BaseDirectory
+
+    format = [
+        '%(levelname)s',
+        '%(process)d',
+        '%(message)s',
+    ]
+    cache = path.join(xdg.BaseDirectory.xdg_cache_home, 'dmedia')
+    if not path.exists(cache):
+        os.makedirs(cache)
+    filename = path.join(cache, namespace + '.log')
+    if path.exists(filename):
+        os.rename(filename, filename + '.previous')
+    logging.basicConfig(
+        filename=filename,
+        filemode='w',
+        level=logging.DEBUG,
+        format='\t'.join(format),
+    )
+    logging.info('dmedia %s, namespace=%r', __version__, namespace)
